"use strict";
/**
 * GitDocumentDB plugin for remote connection using NodeGit
 * Copyright (c) Hidekazu Kubota
 *
 * This source code is licensed under the Mozilla Public License Version 2.0
 * found in the LICENSE file in the root directory of this source tree.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.type = exports.clone = exports.checkPush = exports.checkFetch = exports.sleep = void 0;
const nodegit_1 = __importDefault(require("@sosuisen/nodegit"));
const tslog_1 = require("tslog");
const error_1 = require("./error");
const const_1 = require("./const");
const authentication_1 = require("./authentication");
const net_1 = require("./net");
function sleep(msec) {
    return new Promise(resolve => setTimeout(resolve, msec));
}
exports.sleep = sleep;
const type = 'remote';
exports.type = type;
/**
 * Clone
 *
 * @throws {@link Err.CannotConnectError}
 * @throws {@link Err.CannotCloneRepositoryError}
 */
async function clone(workingDir, remoteOptions, logger) {
    logger !== null && logger !== void 0 ? logger : (logger = new tslog_1.Logger({
        name: 'clone',
        minLevel: 'trace',
        displayDateTime: false,
        displayFunctionName: false,
        displayFilePath: 'hidden',
    }));
    if (remoteOptions !== undefined &&
        remoteOptions.remoteUrl !== undefined &&
        remoteOptions.remoteUrl !== '') {
        /**
         * Retry if network errors.
         */
        let result = {
            ok: false,
        };
        let retry = 0;
        for (; retry < const_1.NETWORK_RETRY; retry++) {
            // eslint-disable-next-line no-await-in-loop
            result = await net_1.checkHTTP(remoteOptions.remoteUrl, const_1.NETWORK_TIMEOUT).catch(err => err);
            if (result.ok) {
                break;
            }
            else {
                logger.debug(`NetworkError in cloning: ${remoteOptions.remoteUrl}, ` + result);
            }
            // eslint-disable-next-line no-await-in-loop
            await sleep(const_1.NETWORK_RETRY_INTERVAL);
        }
        if (!result.ok) {
            // Set retry number for code test
            throw new error_1.Err.CannotConnectError(retry, remoteOptions.remoteUrl, result.toString());
        }
        await nodegit_1.default.Clone.clone(remoteOptions.remoteUrl, workingDir, {
            fetchOpts: {
                callbacks: authentication_1.createCredential(remoteOptions),
            },
        }).catch(err => {
            // Errors except CannotConnectError are handled in combine functions.
            logger.debug(`Error in cloning: ${remoteOptions.remoteUrl}, ` + err);
            throw new error_1.Err.CannotCloneRepositoryError(remoteOptions.remoteUrl);
        });
    }
}
exports.clone = clone;
/**
 * Get or create Git remote named 'origin'
 *
 * (git remote add)
 *
 * @internal
 */
// eslint-disable-next-line complexity
async function _getOrCreateGitRemote(repos, remoteURL) {
    let result;
    // Check if remote repository already exists
    let remote = await nodegit_1.default.Remote.lookup(repos, 'origin').catch(() => { });
    if (remote === undefined) {
        // Add remote repository
        remote = await nodegit_1.default.Remote.create(repos, 'origin', remoteURL);
        result = 'add';
    }
    else if (remote.url() !== remoteURL) {
        nodegit_1.default.Remote.setUrl(repos, 'origin', remoteURL);
        result = 'change';
    }
    else {
        result = 'exist';
    }
    return [result, remote];
}
/**
 * Check connection by FETCH
 *
 * @throws {@link Err.InvalidURLError}
 * @throws {@link Err.RemoteRepositoryNotFoundError}
 * @throws Error (Other errors from NodeGit.Remote#connect())
 *
 * @internal
 */
// eslint-disable-next-line complexity
async function checkFetch(workingDir, options, credentialCallbacks, logger) {
    var _a;
    logger !== null && logger !== void 0 ? logger : (logger = new tslog_1.Logger({
        name: 'clone',
        minLevel: 'trace',
        displayDateTime: false,
        displayFunctionName: false,
        displayFilePath: 'hidden',
    }));
    const repos = await nodegit_1.default.Repository.open(workingDir);
    // Get NodeGit.Remote
    const [gitResult, remote] = await _getOrCreateGitRemote(repos, options.remoteUrl);
    logger.debug('Git remote: ' + gitResult);
    const remoteURL = remote.url();
    const error = String(await remote
        .connect(0 /* FETCH */, credentialCallbacks)
        .catch(err => err));
    await remote.disconnect();
    repos.cleanup();
    // if (error !== 'undefined') console.warn('connect fetch error: ' + error);
    switch (true) {
        case error === 'undefined':
            break;
        case error.startsWith('Error: unsupported URL protocol'):
        case error.startsWith('Error: failed to resolve address'):
        case error.startsWith('Error: failed to send request'):
            throw new error_1.Err.InvalidURLError(remoteURL + ':' + error);
        case error.startsWith('Error: unexpected HTTP status code: 4'): // 401, 404 on Ubuntu
        case error.startsWith('Error: request failed with status code: 4'): // 401, 404 on Windows
        case error.startsWith('Error: Method connect has thrown an error'):
        case error.startsWith('Error: ERROR: Repository not found'):
            // Remote repository does not exist, or you do not have permission to the private repository
            if (((_a = options.connection) === null || _a === void 0 ? void 0 : _a.type) === 'github') {
                return 'not_exist';
            }
            throw new error_1.Err.RemoteRepositoryNotFoundError(remoteURL + ':' + error);
        case error.startsWith('Error: remote credential provider returned an invalid cred type'): // on Ubuntu
        case error.startsWith('Failed to retrieve list of SSH authentication methods'):
        case error.startsWith('Error: too many redirects or authentication replays'):
            throw new error_1.Err.FetchPermissionDeniedError(error);
        default:
            throw new Error(error);
    }
    ;
    return 'exist';
}
exports.checkFetch = checkFetch;
/**
 * Check connection by PUSH
 *
 * @throws {@link Err.InvalidURLError}
 * @throws {@link Err.RemoteRepositoryNotFoundError}
 * @throws {@link Err.PushPermissionDeniedError}
 * @throws Error (Other errors from NodeGit.Remote#connect())
 *
 * @internal
 */
// eslint-disable-next-line complexity
async function checkPush(workingDir, options, credentialCallbacks, logger) {
    logger !== null && logger !== void 0 ? logger : (logger = new tslog_1.Logger({
        name: 'clone',
        minLevel: 'trace',
        displayDateTime: false,
        displayFunctionName: false,
        displayFilePath: 'hidden',
    }));
    const repos = await nodegit_1.default.Repository.open(workingDir);
    // Get NodeGit.Remote
    const [gitResult, remote] = await _getOrCreateGitRemote(repos, options.remoteUrl);
    logger.debug('Git remote: ' + gitResult);
    const error = String(await remote
        .connect(1 /* PUSH */, credentialCallbacks)
        .catch(err => err));
    await remote.disconnect();
    repos.cleanup();
    // if (error !== 'undefined') console.warn('connect push error: ' + error);
    switch (true) {
        case error === 'undefined':
            break;
        case error.startsWith('Error: unsupported URL protocol'):
        case error.startsWith('Error: failed to resolve address'):
        case error.startsWith('Error: failed to send request'):
            throw new error_1.Err.InvalidURLError(options.remoteUrl);
        case error.startsWith('Error: unexpected HTTP status code: 4'): // 401, 404 on Ubuntu
        case error.startsWith('Error: request failed with status code: 4'): // 401, 404 on Windows
        case error.startsWith('Error: Method connect has thrown an error'):
        case error.startsWith('Error: ERROR: Repository not found'): {
            // Remote repository does not exist, or you do not have permission to the private repository
            throw new error_1.Err.RemoteRepositoryNotFoundError(options.remoteUrl);
        }
        // Invalid personal access token
        // Personal access token is read only
        case error.startsWith('Error: remote credential provider returned an invalid cred type'): // on Ubuntu
        case error.startsWith('Error: too many redirects or authentication replays'):
        case error.startsWith('Error: ERROR: Permission to'): {
            throw new error_1.Err.PushPermissionDeniedError(error);
        }
        default:
            throw new Error(error);
    }
    return 'ok';
}
exports.checkPush = checkPush;
//# sourceMappingURL=remote-nodegit.js.map